from rich.columns import Columns
from rich.console import Console, RenderableType
from rich.highlighter import (
    ReprHighlighter,
    _combine_regex,  # pyright: ignore [reportPrivateUsage]
)
from rich.panel import Panel


class DefaultHighlighter(ReprHighlighter):
    highlights = [
        r'(?P<tag_start><)(?P<tag_name>[-\w.:|]*)(?P<tag_contents>[\w\W]*)(?P<tag_end>>)',
        r'(?P<attrib_name>[\w_]{1,50})=(?P<attrib_value>"?[\w_]+"?)?',
        r'(?P<brace>[][{}()])',
        _combine_regex(
            r'(?P<ipv4>[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})',
            r'(?P<ipv6>([A-Fa-f0-9]{1,4}::?){1,7}[A-Fa-f0-9]{1,4})',
            r'(?P<eui64>(?:[0-9A-Fa-f]{1,2}-){7}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{1,2}:){7}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{4}\.){3}[0-9A-Fa-f]{4})',
            r'(?P<eui48>(?:[0-9A-Fa-f]{1,2}-){5}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{1,2}:){5}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{4}\.){2}[0-9A-Fa-f]{4})',
            r'(?P<uuid>[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12})',
            r'(?P<call>[\w.]*?)\(',
            r'\b(?P<bool_true>True)\b|\b(?P<bool_false>False)\b|\b(?P<none>None)\b',
            r'(?P<number_complex>(?<!\w)(?:\-?[0-9]+\.?[0-9]*(?:e[-+]?\d+?)?)(?:[-+](?:[0-9]+\.?[0-9]*(?:e[-+]?\d+)?))?j)',
            r'(?P<number>(?<!\w)\-?[0-9]+\.?[0-9]*(e[-+]?\d+?)?\b|0x[0-9a-fA-F]*)',
            r'(?P<path>\B(/[-\w._+]+)*\/)(?P<filename>[-\w._+]*)?',
            r"(?<![\\\w])(?P<str>b?'''.*?(?<!\\)'''|b?'.*?(?<!\\)'|b?\"\"\".*?(?<!\\)\"\"\"|b?\".*?(?<!\\)\")",
            r'(?P<url>(file|https|http|ws|wss)://[-0-9a-zA-Z$_+!`(),.?/;:&=%#~]*)',
        ),
    ]


highlighter = DefaultHighlighter()
console = Console(highlighter=highlighter)
errconsole = Console(highlighter=highlighter, stderr=True)


def print_project_saved(project: str):
    console.print(f'Project "{project}" saved.')


def print_warning(msg: str):
    errconsole.print(f'[yellow]Warning:[/yellow] {msg}')


def print_error(msg: str):
    errconsole.print(f'[red]Error:[/red] {msg}')


def print_critical(msg: str):
    errconsole.print(f'[violet]Critical:[/violet] {msg}')


def print_tip(msg: str):
    console.print(f'[yellow]Tip:[/yellow] {msg}')


def format_panel_columns(d: dict[str, str]) -> RenderableType:
    columns: list[RenderableType] = []
    for key, output in d.items():
        columns.append(Panel(
            output,
            title=key,
            title_align='left',
            highlight=True,
        ))
    return Columns(columns, expand=True)



import importlib.metadata  # noqa: I001
import typing as t

import typer

from config_keeper.commands.common import autocompletion, helps, sync
from config_keeper.commands.config import cli as config_cli
from config_keeper.commands.paths import cli as paths_cli
from config_keeper.commands.project import cli as project_cli
from config_keeper import config, settings
from config_keeper.output import console

cli = typer.Typer(
    name=settings.EXECUTABLE_NAME,
    help='CLI tool for keeping your personal config files in a repository',
    pretty_exceptions_show_locals=False,
)

cli.add_typer(project_cli, name='project', help='Manage projects.')
cli.add_typer(config_cli, name='config', help='Manage config of this tool.')
cli.add_typer(paths_cli, name='paths', help='Manage project paths.')

ask_help = """
    Ask confirmation before operating.
"""
projects_help = """
    List of project names.
"""
ref_help = """
    Commit sha or branch name to operate with. Only available if specified
    exactly one project. If not given than project branch is used.
"""


@cli.callback(invoke_without_command=True)
def print_version(
    version: t.Annotated[
        bool,
        typer.Option(
            '--version',
            help='Show current version and exit.',
            is_eager=True,
        ),
    ] = False,
):
    if version:
        console.print(
            importlib.metadata.version('config-keeper2'),
            highlight=False,
        )
        raise typer.Exit


@cli.command()
def push(
    projects: t.Annotated[
        t.List[str],  # noqa: UP006
        typer.Argument(
            help=projects_help,
            autocompletion=autocompletion.projects,
        ),
    ],
    ask: t.Annotated[bool, typer.Option(help=ask_help)] = True,
    ref: t.Annotated[
        t.Optional[str],  # noqa: UP007
        typer.Option(help=ref_help),
    ] = None,
    verbose: t.Annotated[
        bool,
        typer.Option('--verbose', '-v', help=helps.verbose),
    ] = False,
):
    """
    Push files or directories of projects to their repositories. This operation
    is NOT atomic (i.e. failing operation for some project does not prevent
    other projects to be processed).
    """

    sync.check_options(projects, ask, ref)
    conf = config.load()
    validator = sync.get_validator('push', conf)
    sync.validate_projects(projects, validator)

    if ref:
        conf['projects'][projects[0]]['branch'] = ref

    sync.handle_push_ask(projects, conf, ask=ask)
    sync.operate('push', projects, conf, verbose)


@cli.command()
def pull(
    projects: t.Annotated[
        t.List[str],  # noqa: UP006
        typer.Argument(
            help=projects_help,
            autocompletion=autocompletion.projects,
        ),
    ],
    ask: t.Annotated[bool, typer.Option(help=ask_help)] = True,
    ref: t.Annotated[
        t.Optional[str],  # noqa: UP007
        typer.Option(help=ref_help),
    ] = None,
    verbose: t.Annotated[
        bool,
        typer.Option('--verbose', '-v', help=helps.verbose),
    ] = False,
):
    """
    Pull all files and directories of projects from their repositories and move
    them to projects' paths with complete overwrite of original files. This
    operation is NOT atomic (i.e. failing operation for some project does not
    prevent other projects to be processed).
    """

    sync.check_options(projects, ask, ref)
    conf = config.load()
    validator = sync.get_validator('pull', conf)
    sync.validate_projects(projects, validator)

    if ref:
        conf['projects'][projects[0]]['branch'] = ref

    sync.handle_pull_ask(projects, conf, ask=ask)
    sync.operate('pull', projects, conf, verbose)



import typing as t

import yaml

from config_keeper import exceptions as exc
from config_keeper import settings


class TProjectBound(t.TypedDict):
    repository: str
    branch: str
    paths: dict[str, str]


TProject = TProjectBound | dict[str, t.Any]


class TConfigBound(t.TypedDict):
    projects: dict[str, TProject]


TConfig = TConfigBound | dict[str, t.Any]


def populate_defaults(config: dict[str, t.Any]):
    for key, type_ in TConfigBound.__annotations__.items():
        config.setdefault(key, type_())


def ensure_exists():
    try:
        settings.CONFIG_FILE.parent.mkdir(parents=True, exist_ok=True)
        settings.CONFIG_FILE.touch()
    except OSError as e:
        raise exc.PublicError(str(e)) from e


def load() -> TConfig:
    ensure_exists()
    try:
        config = yaml.load(
            settings.CONFIG_FILE.read_bytes(),
            Loader=yaml.Loader,
        )
    except yaml.error.YAMLError as e:
        msg = (
            f'{settings.CONFIG_FILE} is not a valid YAML file.\n'
            'Please fix or remove it.'
        )
        tip = (
            'you can use\n'
            f'> {settings.EXECUTABLE_NAME} config validate\n'
            'after.'
        )
        raise exc.InvalidConfigError(msg, tip=tip) from e
    except OSError as e:
        raise exc.PublicError(str(e)) from e

    config = config or {}

    if not isinstance(config, dict):
        msg = (
            f'the root object of {settings.CONFIG_FILE} config must be a map.\n'
            'Please fix or remove config.\n'
        )
        tip = (
            f'you can use\n> {settings.EXECUTABLE_NAME} config validate\n'
            'after.'
        )
        raise exc.InvalidConfigError(msg, tip=tip)

    populate_defaults(config)

    from config_keeper.validation import RootValidator
    validator = RootValidator(config, unknown_param='skip')
    is_valid = validator.validate()
    validator.print_errors()
    if not is_valid:
        raise exc.InvalidConfigError

    return config


def save(config: TConfig):
    raw = yaml.dump(config)
    settings.CONFIG_FILE.write_text(raw)



import functools
import typing as t

import typer

from config_keeper import config
from config_keeper import exceptions as exc


class Context(typer.Context):
    conf: config.TConfig


TCompleteFuncReturn = list[str]
TCompleteFunc = t.Callable[[str, Context], TCompleteFuncReturn]


def _lazy_config(func: TCompleteFunc) -> TCompleteFunc:
    """
    Decorator which loads config only on first call of decorated function. If
    config is invalid then defaults used. Decorated function must take ``ctx``
    argument, in which ``conf`` attribute contains config.
    """
    conf: config.TConfig | None = None

    @functools.wraps(func)
    def wrapper(incomplete: str, ctx: Context) -> TCompleteFuncReturn:
        nonlocal conf
        if conf is None or getattr(wrapper, '__always_load__', False):
            try:
                conf = config.load()
            except exc.InvalidConfigError:
                conf = {}
                config.populate_defaults(conf)
        ctx.conf = conf
        return func(incomplete, ctx)

    return wrapper


@_lazy_config
def project(incomplete: str, ctx: Context) -> TCompleteFuncReturn:
    completions: list[str] = []
    for project in ctx.conf['projects']:
        if project.startswith(incomplete):
            completions.append(project)
    return completions


@_lazy_config
def projects(incomplete: str, ctx: Context) -> TCompleteFuncReturn:
    completions: list[str] = []
    used_projects = ctx.params.get('projects') or ()
    for project in ctx.conf['projects']:
        if project not in used_projects and project.startswith(incomplete):
            completions.append(project)
    return completions


@_lazy_config
def path_names(incomplete: str, ctx: Context) -> TCompleteFuncReturn:
    project = ctx.params.get('project')
    if project:
        paths = ctx.conf['projects'].get(project, {}).get('paths')
        if isinstance(paths, dict):
            all_completions = [str(path) for path in paths]
            used_paths = ctx.params.get('path_names') or ()
            return [
                path
                for path in all_completions
                if path not in used_paths and path.startswith(incomplete)
            ]
    return []



import os
from pathlib import Path

import typer

# paths
_DEFAULT_CONFIG_FILE = (
    Path(typer.get_app_dir('config-keeper', roaming=False)) / 'config.yaml'
)
CONFIG_FILE = Path(os.getenv('CONFIG_KEEPER_CONFIG_FILE', _DEFAULT_CONFIG_FILE))

# etc
EXECUTABLE_NAME = 'config-keeper'



import subprocess
import typing as t

import typer

from config_keeper import config
from config_keeper import exceptions as exc
from config_keeper.output import console, format_panel_columns
from config_keeper.progress import spinner
from config_keeper.sync_handler import SyncHandler
from config_keeper.validation import ProjectValidator, check_if_project_exists

if t.TYPE_CHECKING:
    from rich.progress import TaskID

TOperation = t.Literal['push', 'pull']


def check_options(
    projects: list[str],
    ask: bool = True,
    ref: str | None = None,
):
    if ref is not None and len(projects) > 1:
        msg = '--ref option cannot be used with multiple projects.'
        raise exc.InvalidArgumentError(msg)


def get_validator(
    operation: TOperation,
    conf: config.TConfig,
) -> ProjectValidator:
    if operation == 'push':
        return ProjectValidator(
            conf,
            path_existence='error',
            not_copyable_path='error',
            not_writeable_path='skip',
        )
    return ProjectValidator(
        conf,
        path_existence='skip',
        not_copyable_path='skip',
        not_writeable_path='error',
    )


def validate_projects(
    projects: list[str],
    validator: ProjectValidator,
):
    for project in projects:
        check_if_project_exists(project, validator.conf)

    with spinner() as s:
        prev_task: TaskID | None = None
        for project in projects:
            if prev_task is not None:
                s.stop_task(prev_task)
            prev_task = s.add_task(
                f'Validating project "{project}"...',
                total=None,
            )
            validator.validate(project)

    validator.print_errors()
    if not validator.is_valid:
        raise exc.InvalidConfigError


def handle_push_ask(projects: list[str], conf: config.TConfig, *, ask: bool):
    if not ask:
        return

    console.print('Going to push into following branches:')

    for project in projects:
        console.print(
            f'- "{conf["projects"][project]["branch"]}" at '
            f'{conf["projects"][project]["repository"]} '
            f'(from "{project}")',
        )

    if not typer.confirm('Proceed?', default=True):
        raise typer.Exit


def handle_pull_ask(projects: list[str], conf: config.TConfig, *, ask: bool):
    if not ask:
        return

    console.print('Following paths will most likely be replaced:')

    for project in projects:
        for path in conf['projects'][project]['paths'].values():
            console.print(f'- {path} (from "{project}")')

    if not typer.confirm('Proceed?', default=True):
        raise typer.Exit


def operate(
    operation: TOperation,
    projects: list[str],
    conf: config.TConfig,
    verbose: bool = False,
):
    output: dict[str, str] = {}
    projects_with_errors: list[str] = []

    with spinner() as s:
        prev_task: TaskID | None = None
        for project in projects:
            if prev_task is not None:
                s.stop_task(prev_task)
            prev_task = s.add_task(
                f'Processing project "{project}"...',
                total=None,
            )
            handler = SyncHandler(project, conf, verbose_output=verbose)
            try:
                getattr(handler, operation)()
                output[f'[green]{project}[/green]'] = handler.get_output(
                    verbose,
                )
            except subprocess.CalledProcessError as e:
                projects_with_errors.append(project)
                output[f'[red]{project}[/red]'] = (
                    f'{handler.get_output(verbose)}\n'
                    '[red]'
                    f'{(e.stdout + e.stderr).strip()}'
                    '[/red]'
                ).strip()

    console.print(format_panel_columns(output))

    if projects_with_errors:
        projects_msg = ', '.join(f'"{p}"' for p in projects_with_errors)
        msg = f'operation did not succeeded for {projects_msg}.'
        raise exc.SyncError(msg)

    console.print('Operation [green]successfully[/green] completed.')
